import os
import logging
from datetime import datetime
import synapseclient as syn
from ..core import Utils, AioManager, SynapseProxy
from .file_handle_view import FileHandleView


class Downloader:

    def __init__(self, starting_entity_id, download_path, with_view=False, username=None, password=None):
        self._starting_entity_id = starting_entity_id
        self._download_path = Utils.expand_path(download_path)
        self._with_view = with_view
        self._username = username
        self._password = password

        self.start_time = None
        self.end_time = None

        self._file_handle_view = None
        self.total_files = None
        self.files_processed = 0
        self.has_errors = False

    def start(self):
        self.total_files = None
        self.files_processed = 0
        self.has_errors = False

        if SynapseProxy.logged_in() or SynapseProxy.login(username=self._username, password=self._password):
            AioManager.start(self._startAsync)
        else:
            self.has_errors = True

        self.end_time = datetime.now()
        logging.info('')
        logging.info('Run time: {0}'.format(self.end_time - (self.start_time or datetime.now())))

        if self.has_errors:
            logging.error('Finished with errors. Please see log file.')
        else:
            logging.info('Finished successfully.')

    async def _startAsync(self):
        try:
            Utils.ensure_dirs(self._download_path)

            start_entity = await SynapseProxy.getAsync(self._starting_entity_id, downloadFile=False)

            if type(start_entity) not in [syn.Project, syn.Folder, syn.File]:
                raise Exception('Starting entity must be a Project, Folder, or File.')

            logging.info('Starting Download Entity: {0} ({1})'.format(start_entity.name, start_entity.id))
            logging.info('Downloading to: {0}'.format(self._download_path))

            self._file_handle_view = FileHandleView(start_entity)

            if self._with_view:
                await self._file_handle_view.load()
                self.total_files = len(self._file_handle_view)
                logging.info('Total files: {0}'.format(self.total_files))

            self.start_time = datetime.now()

            if isinstance(start_entity, syn.File):
                await self._download_file(start_entity.id, self._download_path)
            else:
                await self._download_children(start_entity, self._download_path)
        except Exception as ex:
            logging.exception(ex)
            self.has_errors = True

    async def _download_children(self, parent, local_path):
        syn_folders = []

        async for child in SynapseProxy.Aio.get_children(parent, includeTypes=["folder", "file"]):
            child_id = child.get('id')
            child_name = child.get('name')

            if child.get('type') == 'org.sagebionetworks.repo.model.Folder':
                syn_folders.append({'id': child_id, 'name': child_name, 'local_path': local_path})
            else:
                await self._download_file(child_id, local_path)

        for syn_folder in syn_folders:
            await self._download_folder(syn_folder['id'], syn_folder['name'], syn_folder['local_path'])

    async def _download_folder(self, syn_id, name, local_path):
        full_path = os.path.join(local_path, name)
        logging.info('Folder: {0} -> {1}'.format(full_path.replace(self._download_path, ''), full_path))
        Utils.ensure_dirs(full_path)
        await self._download_children(syn_id, full_path)

    async def _download_file(self, syn_id, local_path):
        try:
            filehandle = await self._file_handle_view.get_filehandle(syn_id)

            filename = filehandle.get('fileHandle').get('fileName')
            remote_md5 = filehandle.get('fileHandle').get('contentMd5')
            content_size = filehandle.get('fileHandle').get('contentSize')

            full_path = os.path.join(local_path, filename)

            progress_msg = str(self.files_processed + 1)
            if self.total_files is not None:
                progress_msg += ' of {0}'.format(self.total_files)

            logging.info(
                'File  : {0} -> {1} [{2}]'.format(full_path.replace(self._download_path, ''), full_path, progress_msg))

            can_download = True

            # Only check the md5 if the file sizes match.
            # This way we can avoid MD5 checking for partial downloads and changed files.
            if os.path.isfile(full_path) and os.path.getsize(full_path) == content_size:
                local_md5 = await Utils.get_md5(full_path)
                if local_md5 == remote_md5:
                    can_download = False
                    logging.info('File is current.')

            if can_download:
                await SynapseProxy.Aio.download_file(syn_id,
                                                     full_path,
                                                     filehandle,
                                                     self._file_handle_view.get_filehandle)
        except Exception as ex:
            self.has_errors = True
            logging.exception(ex)

        self.files_processed += 1



import os
import argparse
import logging
from datetime import datetime
from .download import Downloader
from .core import Utils, SynapseProxy
from .compare.comparer import Comparer


def _start_download(args):
    Downloader(args.entity_id,
               args.download_path,
               with_view=args.with_view,
               username=args.username,
               password=args.password).start()


def _start_compare(args):
    Comparer(args.entity_id,
             args.download_path,
             with_view=args.with_view,
             ignores=args.compare_ignore,
             username=args.username,
             password=args.password).start()


def main(args=None):
    parser = argparse.ArgumentParser()
    parser.add_argument('entity_id',
                        metavar='entity-id',
                        help='The ID of the Synapse entity to download or compare (Project, Folder or File).')

    parser.add_argument('download_path',
                        metavar='download-path',
                        help='The local path to save the files to or to compare.')

    parser.add_argument('-u', '--username',
                        help='Synapse username.',
                        default=None)

    parser.add_argument('-p', '--password',
                        help='Synapse password.',
                        default=None)

    parser.add_argument('-ll', '--log-level',
                        help='Set the logging level.',
                        default='INFO')

    parser.add_argument('-ld', '--log-dir',
                        help='Set the directory where the log file will be written.')

    parser.add_argument('-dt', '--download-timeout',
                        help='Set the maximum time (in seconds) a file can download before it is canceled.',
                        type=int,
                        default=SynapseProxy.Aio.FILE_DOWNLOAD_TIMEOUT)

    parser.add_argument('-w', '--with-view',
                        help='Use an entity view for loading file info. Fastest for large projects.',
                        default=False,
                        action='store_true')

    parser.add_argument('-wc', '--with-compare',
                        help='Run the comparison after downloading everything.',
                        default=False,
                        action='store_true')

    # Comparing.
    parser.add_argument('-c', '--compare',
                        help='Compare a local directory against a remote project or folder.',
                        default=False,
                        action='store_true')

    parser.add_argument('-ci', '--compare-ignore',
                        help='Path to directories or files to ignore when comparing.',
                        action='append',
                        nargs='?')

    args = parser.parse_args(args)

    log_level = getattr(logging, args.log_level.upper())

    timestamp = datetime.now().strftime("%Y%m%d%H%M%S%f")
    log_filename = '{0}.log'.format(timestamp)

    if args.log_dir:
        log_filename = os.path.join(Utils.expand_path(args.log_dir), log_filename)
    else:
        log_filename = os.path.join(Utils.app_log_dir(), log_filename)

    Utils.ensure_dirs(os.path.dirname(log_filename))

    logging.basicConfig(
        filename=log_filename,
        filemode='w',
        format='%(asctime)s %(levelname)s: %(message)s',
        level=log_level
    )

    # Add console logging.
    console = logging.StreamHandler()
    console.setLevel(log_level)
    console.setFormatter(logging.Formatter('%(message)s'))
    logging.getLogger().addHandler(console)

    print('Logging output to: {0}'.format(log_filename))

    if args.download_timeout != SynapseProxy.Aio.FILE_DOWNLOAD_TIMEOUT:
        SynapseProxy.Aio.FILE_DOWNLOAD_TIMEOUT = args.download_timeout
        logging.info('Download timeout set to: {0}'.format(SynapseProxy.Aio.FILE_DOWNLOAD_TIMEOUT))

    if args.compare:
        _start_compare(args)
    else:
        _start_download(args)
        if args.with_compare:
            _start_compare(args)

    print('Output logged to: {0}'.format(log_filename))


if __name__ == "__main__":
    main()



import os
import logging
import getpass
import asyncio
import aiohttp
import aiofiles
import random
import hashlib
import synapseclient as syn
import synapseclient.utils as syn_utils
from functools import partial
from .utils import Utils
from .aio_manager import AioManager


class SynapseProxy:
    _synapse_client = None

    @classmethod
    def login(cls, username=None, password=None):
        username = username or os.getenv('SYNAPSE_USERNAME')
        password = password or os.getenv('SYNAPSE_PASSWORD')

        if not username:
            username = input('Synapse username: ')

        if not password:
            password = getpass.getpass(prompt='Synapse password: ')

        logging.info('Logging into Synapse as: {0}'.format(username))
        try:
            # Disable the synapseclient progress output.
            syn.utils.printTransferProgress = lambda *a, **k: None

            cls._synapse_client = syn.Synapse(skip_checks=True)
            cls._synapse_client.login(username, password, silent=True)
        except Exception as ex:
            cls._synapse_client = None
            logging.error('Synapse login failed: {0}'.format(str(ex)))

        return cls._synapse_client is not None

    @classmethod
    def logged_in(cls):
        return cls._synapse_client is not None

    @classmethod
    def client(cls):
        if not cls._synapse_client:
            cls.login()
        return cls._synapse_client

    @classmethod
    def store(cls, obj, **kwargs):
        return cls.client().store(obj, **kwargs)

    @classmethod
    async def storeAsync(cls, obj, **kwargs):
        args = partial(cls.store, obj=obj, **kwargs)
        return await asyncio.get_running_loop().run_in_executor(None, args)

    @classmethod
    def get(cls, entity, **kwargs):
        return cls.client().get(entity, **kwargs)

    @classmethod
    async def getAsync(cls, entity, **kwargs):
        args = partial(cls.get, entity=entity, **kwargs)
        return await asyncio.get_running_loop().run_in_executor(None, args)

    @classmethod
    def getChildren(cls, parent, **kwargs):
        return list(cls.client().getChildren(parent, **kwargs))

    @classmethod
    async def getChildrenAsync(cls, parent, **kwargs):
        args = partial(cls.getChildren, parent=parent, **kwargs)
        return await asyncio.get_running_loop().run_in_executor(None, args)

    @classmethod
    def tableQuery(cls, query, resultsAs="csv", **kwargs):
        return cls.client().tableQuery(query=query, resultsAs=resultsAs, **kwargs)

    @classmethod
    async def tableQueryAsync(cls, query, resultsAs="csv", **kwargs):
        args = partial(cls.tableQuery, query=query, resultsAs=resultsAs, **kwargs)
        return await asyncio.get_running_loop().run_in_executor(None, args)

    @classmethod
    def delete(cls, obj, version=None):
        return cls.client().delete(obj, version=version)

    @classmethod
    async def deleteAsync(cls, obj, version=None):
        args = partial(cls.delete, obj=obj, version=version)
        return await asyncio.get_running_loop().run_in_executor(None, args)

    class Aio:
        # File downloads have a max of 1 hour to download.
        FILE_DOWNLOAD_TIMEOUT = 60 * 60

        @classmethod
        async def rest_post(cls, url, endpoint=None, headers=None, body=None):
            max_attempts = 3
            attempt_number = 0

            while True:
                try:
                    uri, headers = SynapseProxy.client()._build_uri_and_headers(url, endpoint=endpoint, headers=headers)

                    if 'signature' in headers and isinstance(headers['signature'], bytes):
                        headers['signature'] = headers['signature'].decode("utf-8")

                    async with AioManager.AIOSESSION.post(uri, headers=headers, json=body) as response:
                        return await response.json()
                except Exception as ex:
                    logging.exception(ex)
                    attempt_number += 1
                    if attempt_number < max_attempts:
                        sleep_time = random.randint(1, 5)
                        logging.info('  Retrying POST in: {0}'.format(sleep_time))
                        await asyncio.sleep(sleep_time)
                    else:
                        logging.error('  Failed POST: {0}'.format(url))
                        raise

        @classmethod
        async def download_file(cls, syn_id, local_path, filehandle, filehandle_func):
            # TODO: Add resume ability for downloads.
            max_attempts = 3
            attempt_number = 0

            while True:
                try:
                    if attempt_number > 0:
                        # Get a new filehandle for retries so the preSignedURL is fresh.
                        filehandle = await filehandle_func(syn_id)

                    url = filehandle.get('preSignedURL')
                    remote_md5 = filehandle.get('fileHandle').get('contentMd5')
                    total_size = filehandle.get('fileHandle').get('contentSize')

                    mb_total_size = Utils.pretty_size(total_size)
                    timeout = aiohttp.ClientTimeout(total=cls.FILE_DOWNLOAD_TIMEOUT)

                    async with AioManager.AIOSESSION.get(url, timeout=timeout) as response:
                        async with aiofiles.open(local_path, mode='wb') as fd:
                            bytes_read = 0
                            download_md5 = hashlib.md5()
                            while True:
                                chunk = await response.content.read(Utils.CHUNK_SIZE)
                                if not chunk:
                                    break
                                bytes_read += len(chunk)
                                Utils.print_inplace(
                                    'Saving {0} of {1}'.format(Utils.pretty_size(bytes_read), mb_total_size))

                                await fd.write(chunk)
                                download_md5.update(chunk)

                            Utils.print_inplace('')
                            logging.info('Saved {0}'.format(Utils.pretty_size(bytes_read)))

                            if bytes_read != total_size:
                                raise Exception(
                                    'Bytes downloaded: {0} does not match remote size: {1}'.format(bytes_read,
                                                                                                   total_size))

                            local_md5 = download_md5.hexdigest()
                            if local_md5 != remote_md5:
                                raise Exception(
                                    'Local MD5: {0} does not match remote MD5: {1}'.format(local_md5, remote_md5))
                            break
                except Exception as ex:
                    logging.exception(ex)
                    attempt_number += 1
                    if attempt_number < max_attempts:
                        sleep_time = random.randint(1, 5)
                        logging.error('  Retrying file in: {0}'.format(sleep_time))
                        await asyncio.sleep(sleep_time)
                    else:
                        logging.error('  Failed to download file: {0}'.format(local_path))
                        raise

        @classmethod
        async def get_children(cls,
                               parent,
                               includeTypes=["folder", "file", "table", "link", "entityview", "dockerrepo"],
                               sortBy="NAME",
                               sortDirection="ASC"):
            parent_id = parent
            if isinstance(parent, str):
                parent_id = parent
            elif isinstance(parent, syn.Entity):
                parent_id = parent.id
            elif isinstance(parent, dict):
                parent_id = parent['id']
            else:
                raise Exception('Invalid parent object: {0}'.format(parent))

            request = {
                'parentId': parent_id,
                'includeTypes': includeTypes,
                'sortBy': sortBy,
                'sortDirection': sortDirection,
                'includeTotalChildCount': True,
                'nextPageToken': None
            }

            response = {"nextPageToken": "first"}
            while response.get('nextPageToken') is not None:
                response = await cls.rest_post('/entity/children', body=request)
                for child in response['page']:
                    yield child
                request['nextPageToken'] = response.get('nextPageToken', None)

        @classmethod
        async def get_file_handle_id(cls, syn_id):
            request = {
                'includeEntity': True,
                'includeAnnotations': False,
                'includePermissions': False,
                'includeEntityPath': False,
                'includeHasChildren': False,
                'includeAccessControlList': False,
                'includeFileHandles': False,
                'includeTableBundle': False,
                'includeRootWikiId': False,
                'includeBenefactorACL': False,
                'includeDOIAssociation': False,
                'includeFileName': False,
                'includeThreadCount': False,
                'includeRestrictionInformation': False
            }

            res = await cls.rest_post('/entity/{0}/bundle2'.format(syn_id), body=request)

            return res.get('entity').get('dataFileHandleId')

        @classmethod
        async def get_filehandle(cls, syn_id, file_handle_id):
            body = {
                'includeFileHandles': True,
                'includePreSignedURLs': True,
                'includePreviewPreSignedURLs': False,
                'requestedFiles': [{
                    'fileHandleId': file_handle_id,
                    'associateObjectId': syn_id,
                    'associateObjectType': 'FileEntity'
                }]
            }

            res = await cls.rest_post('/fileHandle/batch',
                                      endpoint=SynapseProxy.client().fileHandleEndpoint,
                                      body=body)

            return res.get('requestedFiles', [])[0]



import uuid
import logging
import synapseclient as syn
from ..core.synapse_proxy import SynapseProxy


class FileHandleView(dict):
    COL_ID = 'id'
    COL_DATAFILEHANDLEID = 'dataFileHandleId'

    def __init__(self, scope):
        """

        Args:
            scope: The Project, Folder, or File to scope the view to.
        """
        self.scope = scope
        self.view_project = None
        self.view = None

    async def load(self):
        try:
            if isinstance(self.scope, syn.File):
                self._add_item(self.scope.id, self.scope['dataFileHandleId'])
            elif type(self.scope) in [syn.Project, syn.Folder]:
                await self._create()
                logging.info('Querying file view...')
                query = await SynapseProxy.tableQueryAsync('SELECT * FROM {0}'.format(self.view.id))

                id_col = self._get_table_column_index(query.headers, self.COL_ID)
                col_datafilehandleid = self._get_table_column_index(query.headers, self.COL_DATAFILEHANDLEID)

                logging.info('Loading file view...')
                for row in query:
                    self._add_item(row[id_col], row[col_datafilehandleid])
            else:
                raise Exception('Scope entity must be a Project, Folder, or File.')
        except Exception as ex:
            logging.exception(ex)
            raise
        finally:
            await self._delete()

        return self

    async def get(self, syn_id):
        if syn_id not in self:
            self._add_item(syn_id, await SynapseProxy.Aio.get_file_handle_id(syn_id))
        return self[syn_id]

    async def get_filehandle(self, syn_id):
        view_item = await self.get(syn_id)
        return await SynapseProxy.Aio.get_filehandle(syn_id, view_item.get(self.COL_DATAFILEHANDLEID))

    def _add_item(self, id, datafilehandleid):
        self[id] = {
            self.COL_ID: id,
            self.COL_DATAFILEHANDLEID: datafilehandleid
        }

    def _get_table_column_index(self, headers, column_name):
        """Gets the column index for a Synapse Table Column.
        """
        for index, item in enumerate(headers):
            if item.name == column_name:
                return index

    async def _create(self):
        name = '_TEMP_{0}_TEMP_'.format(str(uuid.uuid4()))
        logging.info('Creating file view project: {0}'.format(name))
        self.view_project = await SynapseProxy.storeAsync(syn.Project(name=name))

        logging.info('Creating file view: {0}'.format(name))
        cols = [
            syn.Column(name=self.COL_ID, columnType='ENTITYID'),
            syn.Column(name=self.COL_DATAFILEHANDLEID, columnType='FILEHANDLEID')
        ]
        schema = syn.EntityViewSchema(name=name,
                                      columns=cols,
                                      properties=None,
                                      parent=self.view_project,
                                      scopes=[self.scope],
                                      includeEntityTypes=[syn.EntityViewType.FILE],
                                      addDefaultViewColumns=False,
                                      addAnnotationColumns=False)
        self.view = await SynapseProxy.storeAsync(schema)

    async def _delete(self):
        if self.view_project:
            logging.info('Deleting file view project: {0}'.format(self.view_project.name))
            await SynapseProxy.deleteAsync(self.view_project)



import os
import logging
import synapseclient as syn
from datetime import datetime
from ..core import Utils, SynapseProxy, AioManager
from ..download import FileHandleView


class Comparer:

    def __init__(self, starting_entity_id, local_path, with_view=False, ignores=None, username=None, password=None):
        self._starting_entity_id = starting_entity_id
        self._local_path = Utils.expand_path(local_path)
        self._with_view = with_view
        self._ignores = []
        for ignore in (ignores or []):
            if ignore.lower().startswith('syn'):
                self._ignores.append(ignore.lower())
            else:
                self._ignores.append(Utils.expand_path(ignore))

        self._username = username
        self._password = password

        self._file_handle_view = None

        self.start_time = None
        self.end_time = None
        self.total_remote_files = None
        self.remote_files_processed = 0
        self.has_errors = False

    def start(self):
        self.total_remote_files = None
        self.remote_files_processed = 0
        self.has_errors = False

        if SynapseProxy.logged_in() or SynapseProxy.login(username=self._username, password=self._password):
            AioManager.start(self._startAsync)
        else:
            self.has_errors = True

        self.end_time = datetime.now()
        self._log_info('Run time: {0}'.format(self.end_time - (self.start_time or datetime.now())))

        if self.has_errors:
            logging.error('Finished with errors. See log file.')
        else:
            logging.info('Finished successfully.')

    async def _startAsync(self):
        try:
            start_entity = await SynapseProxy.getAsync(self._starting_entity_id, downloadFile=False)

            if type(start_entity) not in [syn.Project, syn.Folder, syn.File]:
                raise Exception('Starting entity must be a Project, Folder, or File.')

            logging.info('Starting Compare Entity: {0} ({1})'.format(start_entity.name, start_entity.id))
            logging.info('Comparing to: {0}'.format(self._local_path))
            if self._ignores:
                logging.info('Ignoring: {0}'.format(', '.join(self._ignores)))

            self._file_handle_view = FileHandleView(start_entity)

            if self._with_view:
                await self._file_handle_view.load()
                self.total_remote_files = len(self._file_handle_view)
                logging.info('Total Remote files: {0}'.format(self.total_remote_files))

            self.start_time = datetime.now()

            if isinstance(start_entity, syn.File):
                parent = await SynapseProxy.getAsync(start_entity.parentId)
                await self._check_path(parent, self._local_path, remote_file=start_entity)
            else:
                await self._check_path(start_entity, self._local_path)
        except Exception as ex:
            logging.exception(ex)
            self._log_error('Unknown error. See log file.')

    def _log_error(self, *msg):
        self.has_errors = True
        logging.error('\n'.join(msg))

    def _log_info(self, *msg):
        logging.info('\n'.join(msg))

    async def _check_path(self, parent, local_path, remote_file=None):
        if os.path.isdir(local_path):
            remote_file_filename = remote_file.get('_file_handle').get('fileName') if remote_file else None

            local_dirs, local_files = self._get_local_dirs_and_files(local_path, filename=remote_file_filename)
        else:
            local_dirs, local_files = [[], []]

        remote_dirs, remote_files = await self._get_remote_dirs_and_files(parent, remote_file=remote_file)

        for remote_file in remote_files:
            self.remote_files_processed += 1
            progress_msg = str(self.remote_files_processed)
            if self.total_remote_files is not None:
                progress_msg += ' of {0}'.format(self.total_remote_files)

            local_match = self._find_by_name(local_files, remote_file['name'])

            if (remote_file['id'] in self._ignores) or (local_match and local_match.path in self._ignores):
                self._log_info('[SKIPPING REMOTE FILE]',
                               '  REMOTE [ ]: {0}({1})'.format(remote_file['name'], remote_file['id']),
                               '')

                if local_match:
                    self._log_info('[SKIPPING LOCAL FILE]',
                                   '  LOCAL  [ ]: {0}'.format(local_match.path),
                                   '')
                    local_files.remove(local_match)
                continue

            if local_match:
                local_files.remove(local_match)

                self._log_info('[LOCAL FILE FOUND] [{0}]'.format(progress_msg),
                               '  REMOTE [+]: {0}({1})'.format(remote_file['name'], remote_file['id']),
                               '  LOCAL  [+]: {0}'.format(local_match.path),
                               '')

                remote_size = remote_file['content_size']
                local_size = os.path.getsize(local_match.path)
                if local_size != remote_size:
                    self._log_error('[SIZE MISMATCH]',
                                    '  REMOTE [-]: {0}({1}) ({2})'.format(remote_file['name'],
                                                                          remote_file['id'],
                                                                          Utils.pretty_size(remote_size)),
                                    '  LOCAL  [-]: {0} ({1})'.format(local_match.path, Utils.pretty_size(local_size)),
                                    '')
                else:
                    remote_md5 = remote_file['content_md5']
                    local_md5 = await Utils.get_md5(local_match.path)
                    if local_md5 != remote_md5:
                        self._log_error('[MD5 MISMATCH]',
                                        '  REMOTE [-]: {0}({1}) ({2})'.format(remote_file['name'],
                                                                              remote_file['id'],
                                                                              remote_md5),
                                        '  LOCAL  [-]: {0} ({1})'.format(local_match.path, local_md5),
                                        '')

            else:
                self._log_error('[LOCAL FILE NOT FOUND] [{0}]'.format(progress_msg),
                                '  REMOTE [+]: {0}({1})'.format(remote_file['name'], remote_file['id']),
                                '  LOCAL  [-]: {0}'.format(os.path.join(local_path, remote_file['name'])),
                                '')

        for remote_dir in remote_dirs:
            local_match = self._find_by_name(local_dirs, remote_dir['name'])

            if (remote_dir['id'] in self._ignores) or (local_match and local_match.path in self._ignores):
                self._log_info('[SKIPPING REMOTE DIRECTORY]',
                               '  REMOTE [ ]: {0}({1})'.format(remote_dir['name'], remote_dir['id']),
                               '')

                if local_match:
                    self._log_info('[SKIPPING LOCAL DIRECTORY]',
                                   '  LOCAL  [ ]: {0}'.format(local_match.path),
                                   '')
                    local_dirs.remove(local_match)
                continue

            if local_match:
                local_dirs.remove(local_match)

                self._log_info('[LOCAL DIRECTORY FOUND]',
                               '  REMOTE [+]: {0}({1})'.format(remote_dir['name'], remote_dir['id']),
                               '  LOCAL  [+]: {0}'.format(local_match.path),
                               '')

                await self._check_path(remote_dir, os.path.join(local_path, remote_dir['name']))
            else:
                self._log_error('[LOCAL DIRECTORY NOT FOUND]',
                                '  REMOTE [+]: {0}({1})'.format(remote_dir['name'], remote_dir['id']),
                                '  LOCAL  [-]: {0}'.format(os.path.join(local_path, remote_dir['name'])),
                                '')

        for local_file in local_files:
            remote_match = self._find_by_name(remote_files, local_file.name)

            if (local_file.path in self._ignores) or (remote_match and remote_match['id'] in self._ignores):
                self._log_info('[SKIPPING LOCAL FILE]',
                               '  LOCAL  [ ]: {0}'.format(local_file.path),
                               '')

                if remote_match:
                    self._log_info('[SKIPPING REMOTE FILE]',
                                   '  REMOTE [ ]: {0}({1})'.format(remote_match['name'], remote_match['id']),
                                   '')
                    remote_files.remove(remote_match)
                continue

            if remote_match:
                remote_files.remove(remote_match)
                self._log_info('[REMOTE FILE FOUND]',
                               '  LOCAL  [+]: {0}'.format(local_file.path),
                               '  REMOTE [+]: {0}({1})/{2}'.format(parent['name'], parent['id'], local_file.name),
                               '')
            else:
                self._log_error('[REMOTE FILE NOT FOUND]',
                                '  LOCAL  [+]: {0}'.format(local_file.path),
                                '  REMOTE [-]: {0}({1})/{2}'.format(parent['name'], parent['id'], local_file.name),
                                '')

        for local_dir in local_dirs:
            remote_match = self._find_by_name(remote_dirs, local_dir.name)

            if (local_dir.path in self._ignores) or (remote_match and remote_match['id'] in self._ignores):
                self._log_info('[SKIPPING LOCAL DIRECTORY]',
                               '  LOCAL  [ ]: {0}'.format(local_dir.path),
                               '')

                if remote_match:
                    self._log_info('[SKIPPING REMOTE DIRECTORY]',
                                   '  REMOTE [ ]: {0}({1})'.format(remote_match['name'], remote_match['id']),
                                   '')
                    remote_dirs.remove(remote_match)
                continue

            if remote_match:
                remote_dirs.remove(remote_match)
                self._log_info('[REMOTE DIRECTORY FOUND]',
                               '  LOCAL  [+]: {0}'.format(local_dir.path),
                               '  REMOTE [+]: {0}/{1}({2})'.format(parent['name'],
                                                                   remote_match['name'],
                                                                   remote_match['id']),
                               '')

                await self._check_path(remote_match, os.path.join(local_path, remote_match['name']))
            else:
                self._log_error('[REMOTE DIRECTORY NOT FOUND]',
                                '  LOCAL  [+]: {0}'.format(local_dir.path),
                                '  REMOTE [-]: {0}({1})/{2}'.format(parent['name'], parent['id'], local_dir.name),
                                '')

    def _find_by_name(self, _list, name):
        """Finds an item by its name property in a list.

        Args:
            _list: The list (dicts, syn.Files, or syn.Folders) to search.
            name: The name to search by.

        Returns:
            The matching list item or None.

        Raises:
            Exception if more than one item in the list matches the name.
        """

        if len(_list) and isinstance(_list[-1], dict):
            results = [item for item in _list if item['name'] == name]
        else:
            results = [item for item in _list if item.name == name]

        if len(results) > 1:
            raise Exception('More than one item found matching name: {0}'.format(name))
        elif len(results):
            return results[-1]
        else:
            return None

    def _get_local_dirs_and_files(self, local_path, filename=None):
        """Gets all the directories and files in a local path, or a specific file by its name.

        Args:
            local_path: The local path to get files and folders for.
            filename: The name of a single file to return in local_path (if it exists). If present then only
                      this file will be returned in the files list.

        Returns:
            List of directory paths, List of file paths.
        """
        dirs = []
        files = []

        entries = list(os.scandir(local_path))
        for entry in entries:
            # Do not follow any sym links.
            if entry.is_symlink():
                continue

            if filename:
                if entry.is_file() and entry.name == filename:
                    files.append(entry)
                    break
            else:
                if entry.is_dir():
                    dirs.append(entry)
                else:
                    files.append(entry)

        dirs.sort(key=lambda f: f.name)
        files.sort(key=lambda f: f.name)

        return dirs, files

    async def _get_remote_dirs_and_files(self, parent, remote_file=None):
        """Gets all the directories and files in a remote parent (Project or Folder), or a specific remote file.

        Args:
            parent: The remote project or folder to get directories and files for.
            remote_file: The only remote File to return. If present then only this file is returned in the files list.

        Returns:
            List of dicts with folder attributes, List of dicts with file attributes.
        """
        remote_dirs = []
        remote_files = []

        async def _add_child(id, name, is_file=False):
            entity = {
                'id': id,
                'name': name
            }
            if is_file:
                filehandle = await self._file_handle_view.get_filehandle(id)
                entity['name'] = filehandle.get('fileHandle').get('fileName')
                entity['content_md5'] = filehandle.get('fileHandle').get('contentMd5')
                entity['content_size'] = filehandle.get('fileHandle').get('contentSize')
                remote_files.append(entity)
            else:
                remote_dirs.append(entity)

        if remote_file:
            await _add_child(remote_file.id, remote_file.name, is_file=True)
        else:
            async for child in SynapseProxy.Aio.get_children(parent, includeTypes=["folder", "file"]):
                is_file = child.get('type') == 'org.sagebionetworks.repo.model.FileEntity'
                await _add_child(child.get('id'), child.get('name'), is_file=is_file)

        return remote_dirs, remote_files



import asyncio
import aiohttp


class AioManager:
    AIOSESSION = None

    @classmethod
    def start(cls, func, **kwargs):
        asyncio.run(cls._start_async(func, **kwargs))

    @classmethod
    async def _start_async(cls, func, **kwargs):
        try:
            cls.AIOSESSION = aiohttp.ClientSession(raise_for_status=True)
            await func(**kwargs)
        finally:
            if cls.AIOSESSION:
                await cls.AIOSESSION.close()



import os
import hashlib
import aiofiles
import math
import pathlib


class Utils:
    KB = 1024
    MB = KB * KB
    CHUNK_SIZE = 10 * MB

    @staticmethod
    def app_dir():
        """Gets the application's primary directory for the current user.

        Returns:
            Absolute path to the directory.
        """
        return os.path.join(pathlib.Path.home(), '.syntools')

    @staticmethod
    def app_log_dir():
        """Gets the applications primary log directory for the current user.

        Returns:
            Absolute path to the directory.
        """
        return os.path.join(Utils.app_dir(), 'logs')

    @staticmethod
    def expand_path(local_path):
        var_path = os.path.expandvars(local_path)
        expanded_path = os.path.expanduser(var_path)
        return os.path.abspath(expanded_path)

    @staticmethod
    def ensure_dirs(local_path):
        """Ensures the directories in local_path exist.

        Args:
            local_path: The local path to ensure.

        Returns:
            None
        """
        if not os.path.isdir(local_path):
            os.makedirs(local_path)

    @staticmethod
    def split_chunk(list, chunk_size):
        """Yield successive n-sized chunks from a list.

        Args:
            list: The list to chunk.
            chunk_size: The max chunk size.

        Returns:
            List of lists.
        """
        for i in range(0, len(list), chunk_size):
            yield list[i:i + chunk_size]

    # Holds the last string that was printed
    __last_print_inplace_len = 0

    @staticmethod
    def print_inplace(msg):
        # Clear the line. Using this method so it works on Windows too.
        print(' ' * Utils.__last_print_inplace_len, end='\r')
        print(msg, end='\r')
        Utils.__last_print_inplace_len = len(msg)

    # Hold the names for pretty printing file sizes.
    PRETTY_SIZE_NAMES = ("Bytes", "KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB")

    @staticmethod
    def pretty_size(size):
        if size > 0:
            i = int(math.floor(math.log(size, 1024)))
            p = math.pow(1024, i)
            s = round(size / p, 2)
        else:
            i = 0
            s = 0
        return '{0} {1}'.format(s, Utils.PRETTY_SIZE_NAMES[i])

    @staticmethod
    async def get_md5(local_path):
        md5 = hashlib.md5()
        async with aiofiles.open(local_path, mode='rb') as fd:
            while True:
                chunk = await fd.read(Utils.CHUNK_SIZE)
                if not chunk:
                    break
                md5.update(chunk)
        return md5.hexdigest()



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="synapse-downloader",
    version="0.0.1",
    author="Patrick Stout",
    author_email="pstout@prevagroup.com",
    license="Apache2",
    description="Utility for downloading large datasets from Synapse.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/ki-tools/synapse-downloader",
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    classifiers=(
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ),
    entry_points={
        'console_scripts': [
            "synapse-downloader = synapse_downloader.cli:main"
        ]
    },
    install_requires=[
        "synapseclient>=1.9.2,<2.0.0",
        "aiohttp",
        "aiofiles"
    ]
)
